/* Port forwarder for gdb testsuite.
   Copyright 2004 Free Software Foundation, Inc.
   Contributed by Nathan Sidwell <nathan@codesourcery.com>.
   
   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

/* Usage
     forwarder <gdb-port> <sim-port> <cmd-and-args>

     For each connection made on <gdb-port> a new instance of
     <cmd-and-args> is spawned.  That should connect to <sim-port>,
     and then data is forwarded between the two connections.

     Use in combination with the [undocumented] rdi-stub -c option.
*/

#include <malloc.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <netinet/tcp.h>

#define BUFFER_SIZE 1024

struct forward
{
  int gdb_socket;
  char *data_for_gdb;
  unsigned len_for_gdb;
  unsigned read_for_gdb;
  
  int sim_socket;
  char *data_for_sim;
  unsigned len_for_sim;
  unsigned read_for_sim;

  pid_t sim_pid;
};

int main (int argc, char ** argv)
{
  int num_forwards = 0;
  int alloc_forwards = 10;
  struct forward *forwards = malloc (alloc_forwards * sizeof (*forwards));
  int gdb_server;
  int sim_server;
  struct sockaddr_in addr;
  int gdb_port, sim_port;
  char *pos;
  int waiters = 0;
  int yes = 1;

  if (argc < 3)
    {
    usage:
      fprintf (stderr, "Usage: %s <gdb-port> <sim-port> <cmd-and-args>\n",
	       argv[0]);
      exit (1);
    }
  gdb_port = strtol (argv[1], &pos, 0);
  if (*pos)
    goto usage;
  sim_port = strtol (argv[2], &pos, 0);
  if (*pos)
    goto usage;
  
  printf ("Forwarding gdb from port %d to simulator on port %d\n",
	  gdb_port, sim_port);
  
  gdb_server = socket (PF_INET, SOCK_STREAM, 0);
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl (INADDR_LOOPBACK);
  addr.sin_port = htons (gdb_port);
  if (gdb_server < 0
      || setsockopt (gdb_server, SOL_SOCKET, SO_REUSEADDR,
		     (void *)&yes, sizeof (yes))
      || bind (gdb_server, (struct sockaddr *)&addr, sizeof (addr))
      || listen (gdb_server, 5))
    {
      fprintf (stderr, "%s:failed to create/bind/listen gdb server port\n",
	       argv[0]);
      exit (1);
    }
  
  sim_server = socket (PF_INET, SOCK_STREAM, 0);
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl (INADDR_LOOPBACK);
  addr.sin_port = htons (sim_port);
  if (sim_server < 0
      || setsockopt (sim_server, SOL_SOCKET, SO_REUSEADDR,
		     (void *)&yes, sizeof (yes))
      || bind (sim_server, (struct sockaddr *)&addr, sizeof (addr))
      || listen (sim_server, 5))
    {
      fprintf (stderr, "%s:failed to create/bind/listen sim server port\n",
	       argv[0]);
      exit (1);
    }
  
  while (1)
    {
      fd_set reads;
      fd_set writes;
      unsigned ix;
      struct forward *ptr;
      int highest = 0;
      struct timeval timeout;
      int ready;
      int nd;

      FD_ZERO (&reads);
      FD_ZERO (&writes);
      
      if (highest < gdb_server)
	highest = gdb_server;
      FD_SET (gdb_server, &reads);
      
      if (highest < sim_server)
	highest = sim_server;
      FD_SET (sim_server, &reads);
      
      for (ix = num_forwards, ptr = forwards; ix--; ptr++)
	if (ptr->sim_socket)
	  {
	    if (ptr->len_for_sim)
	      {
		if (highest < ptr->sim_socket)
		  highest = ptr->sim_socket;
		FD_SET (ptr->sim_socket, &writes);
	      }
	    else
	      {
		if (highest < ptr->gdb_socket)
		  highest = ptr->gdb_socket;
		FD_SET (ptr->gdb_socket, &reads);
	      }

	    if (ptr->len_for_gdb)
	      {
		if (highest < ptr->gdb_socket)
		  highest = ptr->gdb_socket;
		FD_SET (ptr->gdb_socket, &writes);
	      }
	    else
	      {
		if (highest < ptr->sim_socket)
		  highest = ptr->sim_socket;
		FD_SET (ptr->sim_socket, &reads);
	      }
	  }
      timeout.tv_sec = 10;
      timeout.tv_usec = 0;
      
      ready = select (highest + 1, &reads, &writes, NULL,
		      waiters ? &timeout : NULL);

      if (ready < 0)
	{
	  fprintf (stderr, "%s: something wierd happened\n", argv[0]);
	  exit (1);
	}

      if (!ready)
	{
	  for (ix = num_forwards, ptr = forwards; ix--; )
	    if (!ptr->sim_socket)
	      {
		close (ptr->gdb_socket);
		waiters--;
		memmove (ptr, ptr + 1, ix * sizeof (*ptr));
		break;
	      }
	    else
	      ptr++;
	}

      if (FD_ISSET (gdb_server, &reads))
	{
	  /* A new gdb_connection.  */
	  int gdb;
	  pid_t child;

	  if (num_forwards == alloc_forwards)
	    {
	      alloc_forwards *= 2;

	      forwards = realloc (forwards,
				  alloc_forwards * sizeof (*forwards));
	    }
	  gdb = accept (gdb_server, NULL, NULL);
	  if (gdb < 0)
	    goto failed;

	  /* Tell TCP not to delay small packets.  This greatly speeds up
	     interactive response. */
	  nd = 1;
	  setsockopt (gdb, IPPROTO_TCP, TCP_NODELAY,
		      (char *) &nd, sizeof (nd));

	  child = fork ();
	  if (!child)
	    {
	      execvp (argv[3], &argv[3]);
	      printf ("failed to exec %s\n", argv[3]);
	      exit (1);
	    }
	  else if (child < 0)
	    {
	      close (gdb);
	    }
	  else
	    {
	      ptr = &forwards[num_forwards++];
	      ptr->gdb_socket = gdb;
	      ptr->sim_socket = 0;
	      ptr->len_for_gdb = ptr->len_for_sim = 0;
	      ptr->read_for_gdb = ptr->read_for_sim = 0;
	      ptr->data_for_gdb = ptr->data_for_sim = NULL;
	      ptr->sim_pid = child;
	      waiters++;
	    }
	failed:;
	}

      if (FD_ISSET (sim_server, &reads))
	{
	  int sim = accept (sim_server, NULL, NULL);

	  if (sim >= 0)
	    {
	      /* Tell TCP not to delay small packets.  This greatly speeds up
		 interactive response. */
	      nd = 1;
	      setsockopt (sim, IPPROTO_TCP, TCP_NODELAY,
			  (char *) &nd, sizeof (nd));

	      if (!waiters)
		exit (0);
	      
	      for (ix = num_forwards, ptr = forwards; ix--; )
		if (!ptr->sim_socket)
		  {
		    waiters--;
		    ptr->sim_socket = sim;
		    ptr->data_for_sim = malloc (BUFFER_SIZE);
		    ptr->data_for_gdb = malloc (BUFFER_SIZE);
		    goto found;
		  }
	      close (sim);
	    found:;
	    }
	}
      
      for (ix = num_forwards, ptr = forwards; ix--; )
	{
	  int bytes;
	  int closed = 0;
	  
	  if (FD_ISSET (ptr->sim_socket, &reads))
	    {
	      bytes = read (ptr->sim_socket, ptr->data_for_gdb, BUFFER_SIZE);
	      ptr->read_for_gdb = 0;
	      ptr->len_for_gdb = bytes;
	      if (bytes <= 0)
		closed = 1;
	    }

	  if (FD_ISSET (ptr->sim_socket, &writes))
	    {
	      bytes = write (ptr->sim_socket,
			     ptr->data_for_sim + ptr->read_for_sim,
			     ptr->len_for_sim);
	      ptr->read_for_sim += bytes;
	      ptr->len_for_sim -= bytes;
	      if (bytes <= 0)
		closed = 1;
	    }
	  
	  if (FD_ISSET (ptr->gdb_socket, &reads))
	    {
	      bytes = read (ptr->gdb_socket, ptr->data_for_sim, BUFFER_SIZE);
	      ptr->read_for_sim = 0;
	      ptr->len_for_sim = bytes;
	      if (bytes <= 0)
		closed = 1;
	    }

	  if (FD_ISSET (ptr->gdb_socket, &writes))
	    {
	      bytes = write (ptr->gdb_socket,
			     ptr->data_for_gdb + ptr->read_for_gdb,
			     ptr->len_for_gdb);
	      ptr->read_for_gdb += bytes;
	      ptr->len_for_gdb -= bytes;
	      if (bytes <= 0)
		closed = 1;
	    }
	  
	  if (closed)
	    {
	      shutdown (ptr->sim_socket, SHUT_RDWR);
	      waitpid (ptr->sim_pid, NULL, 0);
	      close (ptr->gdb_socket);
	      close (ptr->sim_socket);
	      free (ptr->data_for_sim);
	      free (ptr->data_for_gdb);
	      memmove (ptr, ptr + 1, ix * sizeof (*ptr));
	      num_forwards--;
	    }
	  else
	    ptr++;
	}
    }
}
